@@ -97,7 +97,8 @@ static gboolean pref_dialog_done = FALSE;
 static GRand *rgen;
 
 static GtkWidget *draw_area;
-static GtkWidget *app, *headerbar, *pref_dialog, *gridframe;
+static GtkWidget *app, *headerbar, *pref_dialog, *gridframe, *restart_game_dialog;
+static GtkWidget *size_radio_s, *size_radio_m, *size_radio_l;
 
 static gint window_width = 0, window_height = 0;
 static gboolean window_is_fullscreen = FALSE, window_is_maximized = FALSE;
@@ -108,15 +109,12 @@ static field_props field[MAXFIELDSIZE * MAXFIELDSIZE];
 static GamesPreimage *ball_preimage = NULL;
 /* The tile images with balls rendered on them. */
 static cairo_surface_t *ball_surface = NULL;
-/* The balls rendered to a size appropriate for the preview. */
-static cairo_surface_t *preview_surfaces[7] = { NULL, NULL, NULL, NULL, NULL, NULL, NULL };
 
 static GtkImage* preview_images[MAXNPIECES];
 static GdkPixbuf* preview_pixbufs[MAXNPIECES];
 
 /* A cairo_surface_t of a blank tile. */
 static cairo_surface_t *blank_surface = NULL;
-static cairo_surface_t *blank_preview_surface = NULL;
 
 static GamesFileList *theme_file_list = NULL;
 
@@ -261,6 +259,9 @@ refresh_pixmaps (void)
 
   gdk_cairo_set_source_pixbuf (cr, ball_pixbuf, 0, 0);
   cairo_mask (cr, cairo_get_source (cr));
+  g_object_unref (ball_pixbuf);
+
+  cairo_destroy (cr);
 
   cr_blank = cairo_create (blank_surface);
   gdk_cairo_set_source_rgba (cr_blank, &backgnd.color);
@@ -268,8 +269,6 @@ refresh_pixmaps (void)
   cairo_rectangle (cr_blank, 0, 0, boxsize, boxsize);
   cairo_fill (cr_blank);
 
-  g_object_unref (ball_pixbuf);
-  cairo_destroy (cr);
   cairo_destroy (cr_blank);
 }
 
@@ -283,6 +282,9 @@ refresh_preview_surfaces (void)
   GdkRGBA bg;
   cairo_t *cr;
   GdkRectangle preview_rect;
+  cairo_surface_t *blank_preview_surface = NULL;
+  /* The balls rendered to a size appropriate for the preview. */
+  cairo_surface_t *preview_surface = NULL;
 
   context = gtk_widget_get_style_context (widget);
   gtk_style_context_get_background_color (context, GTK_STATE_FLAG_NORMAL, &bg);
@@ -301,10 +303,6 @@ refresh_preview_surfaces (void)
    * set them as the background for each widget in the preview array.
    * This code assumes that each preview window is identical. */
 
-  for (i = 0; i < 7; i++)
-    if (preview_surfaces[i])
-      cairo_surface_destroy (preview_surfaces[i]);
-
   if (ball_preimage)
     scaled = games_preimage_render (ball_preimage, 4 * PREVIEW_IMAGE_WIDTH,
                                     7 * PREVIEW_IMAGE_HEIGHT);
@@ -316,10 +314,10 @@ refresh_preview_surfaces (void)
   }
 
   for (i = 0; i < 7; i++) {
-    preview_surfaces[i] = gdk_window_create_similar_image_surface (gtk_widget_get_window (widget),
-                                                                   CAIRO_FORMAT_ARGB32,
-                                                                   PREVIEW_IMAGE_WIDTH, PREVIEW_IMAGE_HEIGHT, 1);
-    cr = cairo_create (preview_surfaces[i]);
+    preview_surface = gdk_window_create_similar_image_surface (gtk_widget_get_window (widget),
+                                                               CAIRO_FORMAT_ARGB32,
+                                                               PREVIEW_IMAGE_WIDTH, PREVIEW_IMAGE_HEIGHT, 1);
+    cr = cairo_create (preview_surface);
     gdk_cairo_set_source_rgba (cr, &bg);
     gdk_cairo_rectangle (cr, &preview_rect);
     cairo_fill (cr);
@@ -330,15 +328,13 @@ refresh_preview_surfaces (void)
     if (preview_pixbufs[i])
       g_object_unref (preview_pixbufs[i]);
 
-    preview_pixbufs[i] = gdk_pixbuf_get_from_surface (preview_surfaces[i], 0, 0,
+    preview_pixbufs[i] = gdk_pixbuf_get_from_surface (preview_surface, 0, 0,
                                                       PREVIEW_IMAGE_WIDTH, PREVIEW_IMAGE_HEIGHT);
 
     cairo_destroy (cr);
+    cairo_surface_destroy (preview_surface);
   }
 
-  if (blank_preview_surface)
-    cairo_surface_destroy (blank_preview_surface);
-
   blank_preview_surface = gdk_window_create_similar_surface (gtk_widget_get_window (widget),
                                                              CAIRO_CONTENT_COLOR_ALPHA,
                                                              PREVIEW_IMAGE_WIDTH, PREVIEW_IMAGE_HEIGHT);
@@ -347,6 +343,7 @@ refresh_preview_surfaces (void)
   gdk_cairo_rectangle (cr, &preview_rect);
   cairo_fill (cr);
 
+  cairo_surface_destroy (blank_preview_surface);
   cairo_destroy (cr);
   g_object_unref (scaled);
 }
@@ -488,7 +485,8 @@ game_over (void)
   int pos;
 
   set_status_message (_("Game Over!"));
-  pos = games_scores_add_plain_score (highscores, score);
+  if (score > 0)
+    pos = games_scores_add_plain_score (highscores, score);
   show_scores (pos);
 }
 
@@ -1293,8 +1291,47 @@ bg_color_callback (GtkWidget * widget, gpointer data)
 static void
 size_callback (GtkWidget * widget, gpointer data)
 {
-  if (pref_dialog_done)
-    g_settings_set_int (settings, KEY_SIZE, GPOINTER_TO_INT (data));
+  GtkWidget *size_radio, *content_area, *label;
+
+  game_size = g_settings_get_int (settings, KEY_SIZE);
+  if (pref_dialog_done && game_size != data && !restart_game_dialog) {
+    GtkDialogFlags flags = GTK_DIALOG_DESTROY_WITH_PARENT;
+
+    restart_game_dialog = gtk_message_dialog_new (GTK_WINDOW (pref_dialog),
+						  GTK_MESSAGE_WARNING,
+						  flags,
+						  GTK_BUTTONS_NONE,
+						  _("Are you sure you want to restart the game?"));
+
+    gtk_dialog_add_buttons (GTK_DIALOG (restart_game_dialog),
+                            _("_Cancel"), GTK_RESPONSE_CANCEL,
+                            _("_Restart"), GTK_RESPONSE_OK,
+                            NULL);
+
+    gint result = gtk_dialog_run (GTK_DIALOG (restart_game_dialog));
+    gtk_widget_destroy (restart_game_dialog);
+
+    switch (result) {
+    case GTK_RESPONSE_OK:
+      g_settings_set_int (settings, KEY_SIZE, GPOINTER_TO_INT (data));
+      break;
+    case GTK_RESPONSE_CANCEL:
+      switch (game_size) {
+      case SMALL:
+	size_radio = size_radio_s;
+	break;
+      case MEDIUM:
+	size_radio = size_radio_m;
+	break;
+      case LARGE:
+	size_radio = size_radio_l;
+	break;
+      }
+      gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (size_radio), TRUE);
+    }
+
+    restart_game_dialog = NULL;
+  }
 }
 
 static void
@@ -1384,18 +1421,21 @@ game_props_callback (GSimpleAction *action,
                       G_CALLBACK (bg_color_callback), NULL);
 
     size_radio = GTK_WIDGET (gtk_builder_get_object (builder_preferences, "radiobutton_small"));
+    size_radio_s = size_radio;
     g_signal_connect (size_radio, "clicked",
                       G_CALLBACK (size_callback), GINT_TO_POINTER (1));
     if (game_size == SMALL)
       gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (size_radio), TRUE);
 
     size_radio = GTK_WIDGET (gtk_builder_get_object (builder_preferences, "radiobutton_medium"));
+    size_radio_m = size_radio;
     g_signal_connect (size_radio, "clicked",
                       G_CALLBACK (size_callback), GINT_TO_POINTER (2));
     if (game_size == MEDIUM)
       gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (size_radio), TRUE);
 
     size_radio = GTK_WIDGET (gtk_builder_get_object (builder_preferences, "radiobutton_large"));
+    size_radio_l = size_radio;
     g_signal_connect (size_radio, "clicked",
                       G_CALLBACK (size_callback), GINT_TO_POINTER (3));
     if (game_size == LARGE)
@@ -1417,12 +1457,25 @@ game_props_callback (GSimpleAction *action,
   gtk_dialog_run (GTK_DIALOG (pref_dialog));
 }
 
+static void cleanup ()
+{
+  int i = 0;
+
+  for (i = 0; i < G_N_ELEMENTS (preview_images); i++)
+    if (preview_pixbufs[i])
+      g_object_unref (preview_pixbufs[i]);
+
+  g_clear_object (&ball_preimage);
+  g_object_unref (highscores);
+
+}
+
 void
 game_quit_callback (GSimpleAction *action,
                     GVariant *parameter,
                     gpointer user_data)
 {
-  gtk_widget_destroy (app);
+  cleanup ();
 }
 
 void
@@ -1640,6 +1693,7 @@ shutdown_cb (GApplication *application)
   g_settings_set_int (settings, "window-width", window_width);
   g_settings_set_int (settings, "window-height", window_height);
   g_settings_set_boolean (settings, "window-is-maximized", window_is_maximized);
+  cleanup ();
 }
 
 int
@@ -1683,8 +1737,5 @@ main (int argc, char *argv[])
 
   status = g_application_run (G_APPLICATION (application), argc, argv);
 
-  if (ball_preimage)
-    g_object_unref (ball_preimage);
-
   return status;
 }
